
/****************************************************************
********\
 *  deadbrk.c -- iRMX II Deadlock break-time measurement.
 *  A low, medium and high priority task is created. Deadlock
occurs
 *  when the following chronological sequence happens:
 *    (1) low priority task takes exclusive control of a critical
resource
 *    (2) medium or high priority task preempts it.
 *    (3) high priority task requests resource; gets suspended
 *    (4) Medium priority task runs, blocking other two tasks
indefinitely
 *  This situation is handled in iRMX by acquiring a "region"
before
 *  using critical resource, and relinquishing it after use. This
benchmark
 *  measures the overhead involved in "breaking the deadlock".
 *  Compiler: Intel iC-286 V4.1 (LARGE model). Q3 1989, by R. P.
Kar
\****************************************************************
********/

#include <stdio.h>
#include <rmxc.h>

#define MAX_LOOPS 10000
/* Note: This is a CPU-dependent value. It must be set such that
the 
 * execution time for this loop: for (j=0; j < DELAY; j++) spare++
 * is slightly longer than one iRMX sleep period.
 */
#define DELAY     4000

unsigned       el_time, spare, status;
selector       task1_t, task2_t, task3_t, region_t;
unsigned char  pri;
enum           YESNO {NO, YES} dead_brk;
unsigned long  count1, count2, count3, max_loops;
unsigned long  strt_sec, end_sec;
float          deadbrk_time;

/* "union" used to decompose a pointer into segment:offset */
typedef struct {unsigned offset; selector sel;} ptr_s;
union { unsigned *pointer; ptr_s ptr; } ptr_u;

/* Low priority task */
void task1()
{
  unsigned t1_status, j;
  while (1)
  {
    if (count1 == max_loops)
    { printf("deleting task1\n");
      rqdeletetask( (selector)0, &t1_status );            /* delete
self */
    }
    /* Get control over critical region */
    rqreceivecontrol( region_t, &t1_status );

    for (j = 0; j < DELAY; j++) spare++;                   /* delay
loop */

    count1++;
    rqsendcontrol( &t1_status );
  }
}

/* Medium priority task. Only uses CPU time and sleep periodically.
*/
void task2()
{
  unsigned j, t2_status;

  while (1)
  {
    if (count2 == max_loops)
    { printf("deleting task2\n");
      rqdeletetask( (selector)0, &t2_status );           /* delete
self */
    }
    for (j = 0; j < DELAY/4; j++) spare++;                /* delay
loop */

    rqsleep(1, &t2_status);
    count2++;
  }
}

/* High priority task. Potential deadlock when it tries to gain
control
   of the "region" resource, because low-priority task holds region
mostly.
 */
void task3()
{
  unsigned t3_status;

  while (1)
  {
    if (count3 == max_loops)
    { printf("deleting task3\n");
      rqdeletetask( (selector)0, &t3_status );            /* delete
self */
    }
    rqsleep(1, &t3_status);

    /* Ask for control of the region. Relinquish control
immediately after
       receiving it. If task1 is not already holding region, this
should
       take very little time. Otherwise, OS must break deadlock.
     */ 
    if (dead_brk == YES)
    { rqreceivecontrol( region_t, &t3_status );
      rqsendcontrol( &t3_status );
    }
    
    count3++;
  }
}

/********************** Main program ***********************/
main( argc, argv )

unsigned argc;
char *argv[];

{
if (argc > 1)
  max_loops = (unsigned)atoi(argv[1]);
else max_loops = MAX_LOOPS;
   
printf("\nDeadlock break time benchmark\n   %U
loops...\n\n",max_loops);

/* Get priority of main program */
pri = rqgetpriority( (selector)0, &status );

/* Create three tasks. Task1 has lowest priority, task3 has
highest.
 * Measure their execution time WITHOUT deadlocks.
 */
count1 = count2 = count3 = 0;
dead_brk = NO;

task1_t = rqcreatetask( pri+3, task1, ptr_u.ptr.sel, 0L, 512, 0,
&status );
if (status != 0) printf("Create task error\n");

task2_t = rqcreatetask( pri+2, task2, ptr_u.ptr.sel, 0L, 512, 0,
&status );

task3_t = rqcreatetask( pri+1, task3, ptr_u.ptr.sel, 0L, 512, 0,
&status );

strt_sec = rqgettime(&status);                  /* Start of timing
period */

/* Set main program's priority below task 1,2,3 so they run to
completion */
rqsetpriority( (selector)0, pri+4, &status );
while ( (count1 < max_loops) || (count2 < max_loops) || (count3 <
max_loops) )
  rqsleep( 10, &status );

end_sec = rqgettime(&status);                  /* End of timing
period */

el_time = (unsigned)(end_sec - strt_sec);
printf("    Execution time without deadlocks = %u
secs\n\n",el_time);

/* Set main() back to original priority level */
rqsetpriority( (selector)0, pri, &status );

/* Create a "region". To ensure mutually exclusive access to a
critical
   resource a task must acquire the region first */
region_t = rqcreateregion( 1, &status );
if (status != 0) printf("Create region error\n");

count1 = count2 = count3 = 0;
dead_brk = YES;

/* Re-create tasks 1,2,3. Now tasks 1 & 3 will compete for region
*/
task1_t = rqcreatetask( pri+3, task1, ptr_u.ptr.sel, 0L, 512, 0,
&status );
if (status != 0) printf("Create task error\n");

task2_t = rqcreatetask( pri+2, task2, ptr_u.ptr.sel, 0L, 512, 0,
&status );

task3_t = rqcreatetask( pri+1, task3, ptr_u.ptr.sel, 0L, 512, 0,
&status );

strt_sec = rqgettime(&status);                  /* Start of timing
period */

/* Set main program's priority below tasks 1,2,3 so they run to
completion */
rqsetpriority( (selector)0, pri+4, &status);
while ( (count1 < max_loops) || (count2 < max_loops) || (count3 <
max_loops) )
  rqsleep( 10, &status );

end_sec = rqgettime(&status);                  /* End of timing
period */

el_time = (unsigned)(end_sec - strt_sec) - el_time;
printf("    %U deadlock resolutions took %u seconds\n", count3,
el_time);

deadbrk_time = ( (float)el_time/(float)count3 ) * 1000000.0;
printf("    ..... %6.1f microseconds per resolution\n\n",
deadbrk_time);

dqexit(0);
}
